/*
This file implements functions to compute a DCM3 decomposition.
Copyright (C) 2004 The University of Texas at Austin

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<stdio.h>
#include<stdlib.h>
#include<assert.h>
#include<math.h>

#include<constants.h>
#include<phylo_tree.h>
#include<ugraph.h>
#include<dcm3.h>
#include<ugraph_aux_vars.h>

extern int VERBOSE;

void get_dcm3_subsets(phylogenetic_tree T, edge c_edge, int* subsets, int* subsets_, int* separator, int* subset_num) {

  node u, v;
  edge e;
  int subsets_i, separator_i, start_sep, i;
  double cur_dist, shortest_dist;

  for(i = 0; i < T->num_leaves; i++) {
    subsets[i] = -1; 
    separator[i] = -1;
  }
  for(i = 0; i <= T->num_leaves; i++) {
    subsets_[i] = -1; 
  }

  subsets_[0] = 0; *subset_num = 1;
  subsets_i = 0; separator_i = 0; start_sep = 0;

  u = c_edge->u;
  v = c_edge->v;
  
  /*first do subtrees around u*/
  e = u->first_edge;
  while(e){
    if(e != c_edge) {
      cur_dist = 0; shortest_dist = HUGE_VAL;
      get_dcm3_subsets_rec(T, u, e, subsets, separator, &subsets_i, &separator_i, start_sep, cur_dist, &shortest_dist);
      start_sep = separator_i;
      if(VERBOSE>=2)printf("subsets_i = %d\n", subsets_i);
      if(VERBOSE>=2)printf("\n");
      assert(subsets_i <= T->num_leaves);
      subsets_[*subset_num] = subsets_i;
      (*subset_num)++;
    }
    if(e->u == u) e=e->u_next; else e=e->v_next;
  }

  /*then do subtrees around v*/
  e = v->first_edge;
  while(e){
    if(e != c_edge) {
      cur_dist = 0; shortest_dist = HUGE_VAL;
      get_dcm3_subsets_rec(T, v, e, subsets, separator, &subsets_i, &separator_i, start_sep, cur_dist, &shortest_dist);
      start_sep = separator_i;
      if(VERBOSE>=2)printf("subsets_i = %d\n", subsets_i);
      if(VERBOSE>=2)printf("\n");
      assert(subsets_i <= T->num_leaves);
      subsets_[*subset_num] = subsets_i;
      (*subset_num)++;
    }
    if(e->u == v) e=e->u_next; else e=e->v_next;
  }
}

void get_dcm3_subsets_rec(phylogenetic_tree T, node root, edge p_e, int* subsets, int* separator, int* subsets_i, int* separator_i, int start_sep, double cur_dist, double* shortest_dist) {

  node v;
  edge e;
  int i;

  v = opposite(root, p_e);
  /*if we have reached a leaf*/
  if(v->degree == 1) {
    if(cur_dist == *shortest_dist) {
      /*leaves closest to centroid form separator*/
      assert(*separator_i < T->num_leaves);
      separator[*separator_i] = ((node_aux_vars)v->aux_vars)->taxon_id;
      (*separator_i)++;
    }
    else if (cur_dist < *shortest_dist) {
      /*if better leaves are found, then we update the separator 
	and insert previous best leaves into the subset being formed*/
      *shortest_dist = cur_dist;

      for(i = (*separator_i)-1;i >= start_sep && i>-1; i--){
        assert(*subsets_i < T->num_leaves);
	subsets[*subsets_i] = separator[i];
	(*subsets_i)++;
	separator[i] = -1;
      }

      *separator_i = start_sep;
      assert(*separator_i < T->num_leaves);
      separator[*separator_i] = ((node_aux_vars)v->aux_vars)->taxon_id;      
      (*separator_i)++;
    }
    else {
      /*anything not closest to centroid goes into subset*/
      assert(*subsets_i < T->num_leaves);
      subsets[*subsets_i] = ((node_aux_vars)v->aux_vars)->taxon_id;
      (*subsets_i)++;
    }
  }
  /*else keep going deeper*/
  e = v->first_edge;
  while(e){
    if(e != p_e) get_dcm3_subsets_rec(T, v, e, subsets, separator, subsets_i, separator_i,start_sep, cur_dist+((edge_aux_vars)e->aux_vars)->wgt, shortest_dist);
    if(e->u == v) e=e->u_next; else e=e->v_next;
  }
}

edge get_centroid_edge(phylogenetic_tree T){
  
  node v, best_node;
  edge e, centroid_edge;
  double min; short keep_going;
  
  min = HUGE_VAL;
  /*find centroid edge by iterating over all the nodes*/
  best_node = 0;
  for(v = T->G->first_node; v; v = v->next){
    if( abs(((node_aux_vars)v->aux_vars)->taxa_size - T->num_leaves/2) < min){
      best_node = v; 
      min = abs(((node_aux_vars)v->aux_vars)->taxa_size - T->num_leaves/2);
    }
  }
  assert(best_node);
  if(VERBOSE>=3)printf("Best node is %d\n", best_node->id);
  
  e = best_node->first_edge; keep_going = 1;
  while(e && keep_going){
    v = opposite(best_node, e);
    if(((node_aux_vars)v->aux_vars)->taxa_size >= ((node_aux_vars)best_node->aux_vars)->taxa_size){
      centroid_edge = e;
      keep_going = 0;
    }
    if(e->u == best_node)e=e->u_next;
    else e=e->v_next;
  }
  if(keep_going)
    centroid_edge = best_node->first_edge;
  
  return centroid_edge;
}
    
