/*
Copyright (C) 2004 The University of Texas at Austin

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/

/*! \file main.c
\brief Main function to read a newick tree and use it as a guide-tree to produce 
a DCM3 decomposition
*/
#include<string.h>
#include<stdio.h>
#include<stdlib.h>

#include<constants.h>
#include<phylo_tree.h>
#include<ugraph.h>
#include<dcm3.h>

/*! \brief VERBOSE is a global variable used to define the level of 
output. 0 means the output will just be the DCM3 subproblems and higher 
numbers produce more output used for debugging
*/
int VERBOSE;

/*! 
  \brief Returns the DCM3 subproblems as a string where each subset is separated by
  a newline and the taxa in each subset are separated by a tab
  \param -gt Filename containing a newick tree. The tree should not 
  contain any spaces and if it is unweighted then each edge is assigned a weight of 1.
  \param -v Verbose level for output (by default it is 0)
*/
int main(int argc, char* argv[]) {

  char newick_tree[MAX_NEWICK_STR_LEN];
  FILE* fp; int i, j; 
  phylogenetic_tree T; edge c_edge;
  int* subsets;
  int* subsets_;
  int* separator;
  int subset_num, separator_size;
  char* num_to_name[MAX_LEAVES];

  VERBOSE = 0;

  if(argc < 2){
    printf("Type dcm3 -help on how to run\n");
    exit(EXIT_FAILURE);
  }
  for(i = 1; i < argc; i++) {
    /*read in the newick tree*/
    if (!strcmp(argv[i], "-gt")) {
      i++;
      fp = fopen(argv[i], "r");
      if(!fp) {
	printf("Error: could not open filename \"%s\"\n", argv[i]);
	exit(EXIT_FAILURE);
      }
      fscanf(fp, "%s", newick_tree);
      fclose(fp);

      T = new_phylogenetic_tree();
      if(!read_tree(T, num_to_name, newick_tree, 0)) {
	printf("Error: could not read tree %s\n", newick_tree);
	exit(EXIT_FAILURE);
      }
      
    }
    else if (!strcmp(argv[i], "-v")){
      i++;
      sscanf(argv[i], "%d", &VERBOSE);
    }
    else if (!strcmp(argv[i], "-help")){
      printf("Usage: dcm3 -gt <filename containing guide-tree in newick format> (tree must NOT contain any spaces)\n");
      exit(EXIT_FAILURE);
    }
    else {
      printf("Type dcm3 -help on how to run\n");
      exit(EXIT_FAILURE);
    }
  }
  
  if(VERBOSE>=1) printf("Tree has leaves %d\n", T->num_leaves);
  c_edge = get_centroid_edge(T);
  if(VERBOSE>=3) printf("Centroid edge = (%d-%d)\n", c_edge->u->id, c_edge->v->id);
  
  /*this array must have 1 more than number of leaves. it
    stores the index to the start and end of subsets in the subsets array*/
  subsets_ = (int*) malloc((T->num_leaves+1) * sizeof(int));

  subsets = (int*) malloc(T->num_leaves * sizeof(int));
  separator = (int*) malloc(T->num_leaves * sizeof(int));

  get_dcm3_subsets(T, c_edge, subsets, subsets_, separator, &subset_num);

  if(VERBOSE>=1) printf("Number of subsets = %d\n", subset_num-1);
  if(VERBOSE>=1){
    j = 0;
    printf("Separator:");
    for(i = 0; separator[i] != -1 && i < T->num_leaves;i++){
      printf("%s\t", num_to_name[separator[i]]);
      j++;
    }
    printf("\n");
    printf("Separator size = %d\n", j);
    separator_size = j;
  }
  
  for(i = 0; i < subset_num-1;i++){
    if(VERBOSE>=1)printf("%d\n",subsets_[i+1]-subsets_[i]+separator_size);
    for(j = subsets_[i]; j < subsets_[i+1]; j++){
      printf("%s\t", num_to_name[subsets[j]]);
    }
    for(j = 0; separator[j] != -1 && j < T->num_leaves;j++){
      printf("%s\t", num_to_name[separator[j]]);
    }
    printf("\n");
  }
  
  /*delete tree and num_to_name strings*/  
  for(i=0;i<T->num_leaves;i++)free(num_to_name[i]); 
  del_phylogenetic_tree(T);

  /*delete arrays*/
  free(subsets);
  free(subsets_);
  free(separator);
  return 1;
}


