/*! \file phylo_tree.h

Contains definition for a phylogenetic tree structure
*/

#ifndef PHYLO_TREE
#define PHYLO_TREE

#include<ugraph.h>
#include<constants.h>

typedef struct phylogenetic_tree_struct* phylogenetic_tree;
typedef struct node_aux_vars_struct* node_aux_vars;
typedef struct edge_aux_vars_struct* edge_aux_vars;

/*! 
phylogenetic_tree_struct implements a phylogenetic tree structure. It
is built on top of an undirected graph (see ugraph.h) called G. The leaves
are given taxon ids which are integers from 0 to n-1, where n is the
number of leaves. The tree struct contains a mapping from the 
taxon ids to the true taxon names and a variable holding the number of 
leaves.
*/
struct phylogenetic_tree_struct {
 
  ugraph G;
  char* num_to_name[MAX_LEAVES];
  int num_leaves;
};

/* 
Recall that all nodes in the undirected graph (see ugraph.h)  contain a
void* to hold any auxillary variables. For our purposes we want to give
each leaf a taxon id and set this to -1 for internal nodes.

The taxa_size variable contains the number of leaves below a node. This
variable helps us to find the centroid edge quickly. Note that the tree is 
stored in unrooted format, but it is read in as a rooted tree; therefore, 
the parents and children of a node are determined at that time. In the 
unrooted tree, they only help us to find the centroid edge.
*/

struct node_aux_vars_struct {

  int taxon_id;
  int taxa_size;
};

/*
All edges in the undirected graph (see ugraph.h) contain a void* to hold 
any auxillary variables. In our case we just want to remember the weight 
of each edge which is used when computing the separator.
*/
struct edge_aux_vars_struct {

  double wgt;
};

phylogenetic_tree new_phylogenetic_tree();
void del_phylogenetic_tree(phylogenetic_tree T);
void print_phylo_tree(phylogenetic_tree T);

int read_tree(phylogenetic_tree T, char* newick_tree);
int read_tree_rec(phylogenetic_tree T, char** newick_tree, node root);
int read_leaf_name(phylogenetic_tree T, char** newick_tree, node n);

#endif
