/*! \file phylo_tree.h
\brief Contains definition for a phylogenetic tree structure
*/

#ifndef PHYLO_TREE
#define PHYLO_TREE

#include<ugraph.h>
#include<constants.h>

/*! \brief A phylogenetic_tree is just a pointer to a phylogenetic_tree_struct */
typedef struct phylogenetic_tree_struct* phylogenetic_tree;

/*! \brief A node_aux_vars is just a pointer to a node_aux_vars_struct */
typedef struct node_aux_vars_struct* node_aux_vars;

/*! \brief An edge_aux_vars is just a pointer to an edge_aux_vars_struct */
typedef struct edge_aux_vars_struct* edge_aux_vars;

/*! 
\brief Representation of a phylogenetic tree.
*/
struct phylogenetic_tree_struct {
 
  /*!Undirected graph storing the unrooted tree structure*/
  ugraph G;
  
  /*!Mapping from the taxon names to taxon ids used to represent the leaves in the tree*/
  char* num_to_name[MAX_LEAVES];

  /*!Number of leaves in the tree*/
  int num_leaves;
};

/*! 
\brief Holds set of auxillary variables assigned to each node.
Recall that all nodes in the undirected graph (see ugraph.h)  contain a
void* to hold any auxillary variables. 
*/
struct node_aux_vars_struct {
  
  /*!Each leaf is given a taxon id; this variable is set to -1 for internal nodes.*/
  int taxon_id;
  
  /*!
    The taxa_size variable contains the number of leaves below a node. This
    variable helps us to find the centroid edge quickly. Note that the tree is 
    stored in unrooted format, but it is read in as a rooted tree; therefore, 
    the parents and children of a node are determined at that time. The 
    taxa_size variable helps us to find the centroid edge---the algorithm
    to find the centroid iterates over all nodes and finds the one whose 
    taxa_size is closest to half the number of leaves.
  */
  int taxa_size;
};

/*!
\brief Holds set of auxillary variables assigned to each edge.
All edges in the undirected graph (see ugraph.h) contain a void* to hold 
any auxillary variables. 
*/
struct edge_aux_vars_struct {

  /*!Used to store the weight of each edge which is used when computing the separator*/
  double wgt;
};

/*! \brief Create a new phylogenetic tree in constant time.*/
phylogenetic_tree new_phylogenetic_tree();

/*! \brief Delete the phylogenetic tree T in linear time.*/
void del_phylogenetic_tree(phylogenetic_tree T);

/*! \brief Print the adjacency list representation of the phylogenetic tree in
linear time.*/
void print_phylo_tree(phylogenetic_tree T);

/*! \brief Reads in a tree in newick format into an indirected graph in linear time.
The tree is built from bottom up and is read in as a rooted tree. After
the read is complete, the root and any vertices of degree two are deleted.
*/
int read_tree(phylogenetic_tree T, char* newick_tree);

/*! \brief NOT TO BE CALLED BY USER: auxillary function used by read_tree.*/
int read_tree_rec(phylogenetic_tree T, char** newick_tree, node root);

/*! \brief NOT TO BE CALLED BY USER: auxillary function used by read_tree_rec to read in a 
leaf name.*/
int read_leaf_name(phylogenetic_tree T, char** newick_tree, node n);

#endif

