/*! \file ugraph.h

\brief Contains definitions for an undirected graph structure
*/

#ifndef UGRAPH
#define UGRAPH

/*! \brief A node is just a pointer to a node_struct */
typedef struct node_struct* node;

/*! \brief An edge is just a pointer to an edge_struct */
typedef struct edge_struct* edge;

/*! \brief A ugraph is just a pointer to a ugraph_struct */
typedef struct ugraph_struct* ugraph;

/*! 
\brief Representation of a node structure in an undirected graph.
The set of nodes is stored as a doubly linked list.
*/
struct node_struct {

  /*!Previous node in the list of nodes*/ 
  node prev;
  
  /*!Next node in the list of nodes*/
  node next;

  /*!First adjacent edge in the list of edges*/
  edge first_edge;

  /*!Last adjacent edge in the list of edges*/
  edge last_edge;

  /*!Degree of node*/
  int degree;

  /*!Unique ID of node*/
  unsigned id;

  /*!Pointer to auxillary variables, if any; similar to templates*/
  void* aux_vars;
};

/*!
\brief Representation of an edge structure in an undirected graph.
The set of edges is stored as a doubly linked list.
*/
struct edge_struct {

  /*!Node at one endpoint of the edge*/
  node u;

  /*!Node at the other endpoint of the edge*/
  node v;
  
  /*!The previous edge in the list of edges*/
  edge prev;

  /*!The next edge in the list of edges*/
  edge next;
  
  /*!Stores the next edge in the list of adjacent edges to node u*/
  edge u_next;

  /*!Stores the previous edge in the list of adjacent edges to node u*/
  edge u_prev;
  
  /*!Stores the next edge in the list of adjacent edges to node v*/
  edge v_next;

  /*!Stores the previous edge in the list of adjacent edges to node v*/
  edge v_prev;
  
  /*!Unique ID of node*/
  unsigned id;

  /*!Pointer to auxillary variables, if any; similar to templates*/
  void* aux_vars;
};

/*! \brief An undirected graph is represented as two doubly linked lists, 
  one of nodes and the other of edges.*/
struct ugraph_struct {

  /*!First edge in the edge list of the graph*/
  edge first_edge;
  
  /*!Last edge in the edge list of the graph*/
  edge last_edge;

  
  /*!First node in the node list of the graph*/
  node first_node;

  /*!Last node in the node list of the graph*/
  node last_node;
  
  /*!Number of edges*/
  int n_edges;

  /*!Number of nodes*/
  int n_nodes;
  
  /*!Max ID number of edges*/
  unsigned id_edges;

  /*!Max ID number of nodes*/
  unsigned id_nodes;
};

/*! \brief Create a new node in constant time.*/
node new_node(ugraph G, void* aux_vars_);

/*! \brief Delete a node and all edges adjacent to it in O(degree(u)) time.*/
void del_node(ugraph G, node u);

/*! \brief Create a new edge in constant time. */
edge new_edge(ugraph G, node u, node v, void* aux_vars_);

/*! \brief Delete an edge in constant time*/
void del_edge(ugraph G, edge e);

/*! \brief Create a new undirected graph in constant time.*/
ugraph new_graph();

/*! \brief Delete the graph in O(#edges + #nodes) time.*/
void del_graph(ugraph G);

/*! \brief Print the adjacency list representation of a graph in O(#edges + #nodes) time.*/
void print_graph(ugraph G);

/*! \brief Return the node opposite node u on edge e in constant time. */
node opposite(node u, edge e);
#endif












































