/*
This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<constants.h>
#include<phylo_tree.h>
#include<hashtable.h>
#include<assert.h>
#include<scm_merge.h>

int VERBOSE=0;

static void read_chars(FILE* fp, char*** num_to_names, char*** sequences, 
		int* ntax, int* nchars);

int main(int argc, char* argv[]){

  char newick_tree[MAX_NEWICK_STR_LEN];
  char* output_filename; char* out_tree;
  FILE* fp; int i, j, t; 
  phylogenetic_tree* tree_array;
  char** num_to_name;
  char** sequences;
  int ntax, nchars, num_trees, t2s_time, s2t_time, scm_time;
  hashtable name_to_num;
  phylogenetic_tree scm_tree;
  
  /*initialization*/
  num_trees = -1;
  t2s_time=0; s2t_time=0; scm_time=0;
  output_filename = NULL;
  /***********/

  if(argc < 2){
    printf("Type scm_merge -help on how to run\n");
    exit(EXIT_FAILURE);
  }
  for(i = 1; i < argc; i++) {
    /*read in sequences*/
    if (!strcmp(argv[i], "-s")){
      i++;
      fp = fopen(argv[i], "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", argv[i]);
        exit(EXIT_FAILURE);
      }
      read_chars(fp, &num_to_name, &sequences, &ntax, &nchars);
      assert(num_to_name); assert(sequences);
      name_to_num = new_hashtable(ntax);
      for(j=0;j<ntax;j++){
	insert_hashtable(name_to_num, num_to_name[j], j);
      }
      fclose(fp);
    }
    /*read in number of trees*/
    else if (!strcmp(argv[i], "-n")){
      i++;
      sscanf(argv[i], "%d", &num_trees);
    }
    /*read in output filename*/
    else if (!strcmp(argv[i], "-o")){
      output_filename = (char*) malloc(MAX_FILENAME * sizeof(char));
      assert(output_filename);
      i++;
      sscanf(argv[i], "%s", output_filename);
    }
    /*read in the newick trees*/
    else if (!strcmp(argv[i], "-t")) {
      assert(num_trees > 1);
      tree_array = (phylogenetic_tree*) malloc(num_trees * sizeof(phylogenetic_tree));

      i++;
      fp = fopen(argv[i], "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", argv[i]);
        exit(EXIT_FAILURE);
      }
      j = 0;
      while(fscanf(fp, "%s", newick_tree) != EOF){
	assert(name_to_num);
	tree_array[j] = new_phylogenetic_tree();
	read_tree(tree_array[j], num_to_name, newick_tree, name_to_num);
	j++;
      }
      num_trees = j;
      fclose(fp);
    }
    /*VERBOSE variable for debugging*/
    else if (!strcmp(argv[i], "-v")){
      i++;
      sscanf(argv[i], "%d", &VERBOSE);
    }
    /*error*/
    else if (!strcmp(argv[i], "-help")){
      printf("Usage: scm_merge\n\t-n <num of subtrees>\n\t-s <filename containing full dataset on which subtrees were computed (in phylip sequential format)>\n\t-t <filename containing newick trees on each line>\n");
      exit(EXIT_FAILURE);
    }
    /*error*/
    else {
      printf("Type scm_merge -help on how to run\n");
      exit(EXIT_FAILURE);
    }
  }

  convert_tree(tree_array, name_to_num);

  for(i=0;i<num_trees;i++)
    del_phylogenetic_tree(tree_array[i]);
  free(tree_array);

  out_tree=(char*)malloc(MAX_NEWICK_STR_LEN*sizeof(char));
  for(t=0;t<MAX_NEWICK_STR_LEN;t++) out_tree[t] = '\0';

  print_tree(scm_tree, num_to_name, out_tree, 1);
  
  if(output_filename){
    fp = fopen(output_filename, "w");
    if(!fp) {
      printf("Error: could not open filename \"%s\"\n", output_filename);
      exit(EXIT_FAILURE);
    }
    fprintf(fp, "%s\n", out_tree);
    fclose(fp);
  }
  else 
    printf("%s\n", out_tree);
  
  free(out_tree);
  del_phylogenetic_tree(scm_tree);
  for(i=0;i<ntax;i++)free(sequences[i]); free(sequences);
  for(i=0;i<ntax;i++)free(num_to_name[i]); free(num_to_name);
  del_hashtable(name_to_num);
  
  return 1;
}

void read_chars(FILE* fp, char*** num_to_names_, char*** sequences_,
		int* ntax, int* nchars){
  
  int i; char** sequences; char** num_to_names;

  *ntax=0; *nchars=0;
  fscanf(fp, "%d", ntax);
  fscanf(fp, "%d", nchars);
  if(*ntax<=0 || *nchars<=0){
    printf("Error: number of taxa=%d and number of sites=%d must be positive", *ntax, *nchars);
    exit(EXIT_FAILURE);
  }
  
  sequences=(char**) malloc(*ntax * sizeof(char*));
  num_to_names=(char**) malloc(*ntax * sizeof(char*));
  for(i=0;i<*ntax;i++){
    sequences[i] = (char*) malloc((*nchars+1) * sizeof(char));
    num_to_names[i] = (char*) malloc(MAX_NAME_LEN * sizeof(char));
    fscanf(fp, "%s", num_to_names[i]);
    fscanf(fp, "%s", sequences[i]);
  }

  /*
  for(i=0;i<*ntax;i++)
    printf("%s\t%s\n", num_to_names[i], sequences[i]);
  */

  *sequences_ = sequences;
  *num_to_names_ = num_to_names;
}

