/*
This file implements the recursive-iterative-dcm3 program.
Copyright (C) 2004 The University of Texas at Austin

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<constants.h>
#include<phylo_tree.h>
#include<math.h>
#include<assert.h>
#include<sys/types.h>
#include<unistd.h>
#include<signal.h>
#include<recidcm3.h>
#include<time.h>

/*static functions*/
static void read_chars(FILE* fp, char*** num_to_names_, char*** sequences_,
		       int* ntax, int* nchars);
void cleanup(int dummy);

/*************/

/*global*/
int VERBOSE=0; int DEBUG=0;
char recdcm3tree[MAX_FILENAME];
char restricted_gt[MAX_FILENAME];
char subproblemfile[MAX_FILENAME]; 
char subtreefile[MAX_FILENAME]; 
char scorefile[MAX_FILENAME];
char globalsearchtree[MAX_FILENAME];
/********/

int main(int argc, char* argv[]){

  char newick_tree[MAX_NEWICK_STR_LEN];

  FILE* fp; FILE* OUT; phylogenetic_tree T, U;
  double max_subset_size, treescore;
  int num_iterations, ntax, nchars, i, j, maxtime;

  char** sequences;  char** num_to_name;
  hashtable name_to_num;

  char basecmd[MAX_COMMAND]; char globalcmd[MAX_COMMAND]; 
  char buffer[MAX_BUFFER]; char gt_filename[MAX_FILENAME];
  char seq_filename[MAX_FILENAME]; 

  char* output_filename; char* _pid_string_;  

  int_set full_taxa;

  pid_t pid;
  
  time_t starttime; time_t currenttime;

  /************/
  /*initialize*/
  /************/
  starttime=time(NULL);
  signal(SIGTERM, (void (*)(int)) cleanup);
  signal(SIGABRT, (void (*)(int)) cleanup);
  sequences = NULL; T = NULL; U = NULL;
  max_subset_size = HUGE_VAL;
  name_to_num = NULL; 
  OUT=NULL; output_filename=NULL; _pid_string_=NULL;
  num_iterations = 500; maxtime = -1;
  subproblemfile[0] = '\0'; subtreefile[0] = '\0';
  
  strcpy(recdcm3tree, "RECDCM3TREE");
  strcpy(restricted_gt, "RESTRICTEDGT");
  strcpy(subproblemfile,"SUBPROBLEM"); 
  strcpy(subtreefile,"SUBTREE");
  strcpy(scorefile,"ITIMPRTREESCORE");
  strcpy(globalsearchtree, "ITIMPRTREE");
  /************/

  if(argc < 2){
    printf("Type recidcm3 -help for usage\n");
    exit(EXIT_FAILURE);
  }
  for(i = 1; i < argc; i++) {
    /*read in sequences*/
    if (!strcmp(argv[i], "-s")){
      i++; sscanf(argv[i], "%s", seq_filename);
      assert(seq_filename[0] != '-');
      fp = fopen(seq_filename, "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", seq_filename);
        cleanup(1);
        exit(EXIT_FAILURE);
      }
      read_chars(fp, &num_to_name, &sequences, &ntax, &nchars);
      fclose(fp);
      assert(num_to_name); assert(sequences);
      name_to_num = new_hashtable(ntax);
      full_taxa = new_int_set(32);
      for(j=0;j<ntax;j++){
	insert_int_set(full_taxa, j);
	/*printf("insering name %s ", num_to_name[j]);*/
	insert_hashtable(name_to_num, num_to_name[j], j);
	/*printf("at %d\n", lookup_hashtable(name_to_num, num_to_name[j]));*/
      }
    }
    else if (!strcmp(argv[i], "-gt")) {
      i++; sscanf(argv[i], "%s", gt_filename);
      assert(gt_filename[0] != '-');
      fp = fopen(gt_filename, "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", gt_filename);
        cleanup(1);
        exit(EXIT_FAILURE);
      }
      fscanf(fp, "%s", newick_tree);
      fclose(fp);
      
      /*read in guide tree*/
      assert(name_to_num);
      T = new_phylogenetic_tree();
      if(!read_tree(T, num_to_name, newick_tree, name_to_num)) {
	printf("Error: could not read tree %s\n", newick_tree);
        cleanup(1);
	exit(EXIT_FAILURE);
      }
    }
    else if (!strcmp(argv[i], "-o")){
      /*read in output filename and open file for writing*/
      output_filename = (char*) malloc(MAX_FILENAME * sizeof(char));
      i++; sscanf(argv[i], "%s", output_filename);
      assert(output_filename[0] != '-');
      OUT = fopen(output_filename, "w");
      if(!OUT) { 
        printf("Error: could not open filename \"%s\"\n", output_filename); 
        exit(EXIT_FAILURE);
      }  
    }
    else if (!strcmp(argv[i], "-debug")){
      DEBUG = 1;
    }
    else if (!strcmp(argv[i], "-v")){
      i++; sscanf(argv[i], "%s", buffer);
      assert(buffer[0] != '-');
      VERBOSE = atoi(buffer);
    }
    else if (!strcmp(argv[i], "-fs")){
      _pid_string_ = (char*) malloc(MAX_BUFFER * sizeof(char));
      i++; sscanf(argv[i], "%s", _pid_string_);
      assert(_pid_string_[0] != '-');
    }
    else if (!strcmp(argv[i], "-ms")){
      i++; sscanf(argv[i], "%s", buffer);
      assert(buffer[0] != '-');
      max_subset_size = strtod(buffer, 0);
    }
    else if (!strcmp(argv[i], "-it")){
      i++; sscanf(argv[i], "%s", buffer);
      assert(buffer[0] != '-');
      num_iterations = atoi(buffer);
    }
    else if (!strcmp(argv[i], "-mt")){
      i++; sscanf(argv[i], "%s", buffer);
      assert(buffer[0] != '-');
      maxtime = atoi(buffer);
    }
    else if (!strcmp(argv[i], "-bcmd")){
      i++; assert(!strcmp(argv[i], "[")); i++;
      while(i < argc && strcmp(argv[i], "]")){
        strcat(basecmd, argv[i]); i++;
        strcat(basecmd, " "); 
      }
      if(i==argc) { printf("Error: no \']\' found\n"); cleanup(1); exit(EXIT_FAILURE); }
    }
    else if (!strcmp(argv[i], "-gcmd")){
      i++; assert(!strcmp(argv[i], "[")); i++;
      while(i < argc && strcmp(argv[i], "]")){
        strcat(globalcmd, argv[i]); i++;
        strcat(globalcmd, " "); 
      }
      if(i==argc) { printf("Error: no \']\' found\n"); cleanup(1); exit(EXIT_FAILURE); }
    }
    else if (!strcmp(argv[i], "-help")){
      printf("%s", "Usage: recidcm3\n\t(1) -s <filename containing full dataset in phylip sequential format>\n\t(2) -gt <filename containing guide-tree in newick format>\n\t(tree must NOT contain any spaces)\n\t(3) -it <num of iterations> (default=500)\n\t(4) -ms <maximum subset size>(default=DCM3 decomposition)\n\t(5) -o <output filename> (default=screen output)\n\t(6) -fs <file suffix to be appended to all temporary files>\n\t(default=process id number)\n\t(7) -debug (does not delete the temporary files)\n\n\t(8) -bcmd [<system command to compute subtrees>]\n\t(recidcm3 creates temporary files and makes the system call:\n\t\"<bcmd> SUBPROBLEM<fs> RESTRICTEDGT<fs> SUBTREE<fs> <fs>\"\n\n\tFiles used in <bcmd> should be in PATH or working dir.\n\tDescription of parameters to <bcmd>:\n\tSUBPROBLEM<fs> = filename containing subproblem sequences in phylip\n\tRESTRICTEDGT<fs> = filename containing the restricted guide-tree\n\tin newick format\n\tSUBTREE<fs> = filename where the computed subtree\n\tshould be stored by the base program in newick format\n\t<fs> = file suffix, from -fs option, appended to all filenames)\n\n\t(9) -gcmd [<system command for iterative improvement on rec-dcm3 tree>]\n\t(recidcm3 creates temporary files and makes the system call:\n\t\"<gcmd> <s> RECDCM3TREE<fs> ITIMPRTREE<fs> ITIMPRTREESCORE<fs> <fs>\"\n\n\tFiles used in <gcmd> should be in PATH or working dir.\n\tDescription of parameters to <gcmd>:\n\t<s> = filename containing the full set of sequences (see arg 1)\n\tRECDCM3TREE<fs> = filename containing the recursive-dcm3 tree\n\twhich is used as a starting point for the iterative improvement search\n\tITIMPRTREE<fs> = filename where the result of the\n\titerative improvement search should be stored in newick format\n\tITIMPRTREESCORE<fs> = filename where the score of the iterative\n\timprovement search tree should be stored as a number on the first line\n\t<fs> = file suffix, from -fs option, appended to all filenames)\n\n\t(10) -mt <timelimit for recidcm3 in seconds>(default=-1, i.e. infinite)\n");
      exit(EXIT_FAILURE);
    }
    else {
      printf("Type recidcm3 -help for usage\n");
      exit(EXIT_FAILURE);
    }
  }

  if(!_pid_string_) { 
    _pid_string_ = (char*) malloc(MAX_BUFFER * sizeof(char));
    pid=getpid(); sprintf(_pid_string_, "%d", pid); 
  }
  strcat(recdcm3tree, _pid_string_); 
  strcat(subproblemfile, _pid_string_); strcat(subtreefile, _pid_string_); 
  strcat(scorefile, _pid_string_); strcat(restricted_gt, _pid_string_); 
  strcat(globalsearchtree, _pid_string_);

  /*creating base and global command strings*/
  strcat(basecmd, " "); strcat(basecmd, subproblemfile); strcat(basecmd, " "); 
  strcat(basecmd, restricted_gt); strcat(basecmd, " ");
  strcat(basecmd, subtreefile); strcat(basecmd," "); 
  strcat(basecmd,_pid_string_);

  strcat(globalcmd, " "); strcat(globalcmd, seq_filename); strcat(globalcmd, " "); 
  strcat(globalcmd, recdcm3tree); strcat(globalcmd, " ");
  strcat(globalcmd, globalsearchtree); strcat(globalcmd," ");  
  strcat(globalcmd, scorefile); strcat(globalcmd, " ");
  strcat(globalcmd,_pid_string_);
  
  assert(T && sequences);

  /*printing guide-tree to file called RECDCM3TREE*/
  fp = fopen(recdcm3tree, "w");
  if(!fp) { 
    printf("Error: could not open filename \"%s\"\n", recdcm3tree); 
    cleanup(1);
    exit(EXIT_FAILURE);
  } 
  fprintf(fp, "%s", newick_tree); fclose(fp);
  /**********/

  if(OUT){
    fprintf(OUT, "System command for iterative improvement search:\n%s\n\n", globalcmd);
    fprintf(OUT, "System command for base method:\n%s\n\n", basecmd); 
    fprintf(OUT, "Number of iterations=%d, maximum subset size=%.2f, timelimit=%d\n\n", num_iterations, max_subset_size, maxtime);
    fprintf(OUT, "Starting Recursive-Iterative-DCM3...\n");
  }
  else {
    printf("System command for iterative improvement search:\n%s\n\n", globalcmd);
    printf("System command for base method:\n%s\n\n", basecmd); 
    printf("Number of iterations=%d, maximum subset size=%.2f, timelimit=%d\n\n", num_iterations, max_subset_size, maxtime);
    printf("Starting Recursive-Iterative-DCM3...\n");
  } 

  currenttime=time(NULL);
  for(j=0; j<num_iterations && 
	(maxtime == -1 || difftime(currenttime,starttime)<maxtime); j++){
    if(OUT){
      fprintf(OUT, "Iteration %d...\n", j+1);
      fprintf(OUT, "Computing Recursive-DCM3 tree...\n");
    }
    else {
      printf("Iteration %d...\n", j+1);
      printf("Computing Recursive-DCM3 tree...\n");
    }
    U = recursive_dcm3(T, sequences, num_to_name, name_to_num, max_subset_size, basecmd, subproblemfile, restricted_gt, subtreefile, 0);
    assert(U);
    
    currenttime=time(NULL);
    if(OUT) fprintf(OUT, "Total time (hrs): %.5f\n", difftime(currenttime, starttime)/3600);
    else   printf("Total time (hrs): %.5f\n", difftime(currenttime, starttime)/3600);

    for(i=0;i<strlen(newick_tree);i++) newick_tree[i]='\0';
    /*print tree with edge weights*/
    print_tree(U, num_to_name, newick_tree, 1);
    if(OUT) fprintf(OUT, "Rec-DCM3 Tree: %s\n", newick_tree);
    else    printf("Rec-DCM3 Tree: %s\n", newick_tree);

    /*printing Rec-DCM3 tree to file called RECDCM3TREE*/
    fp = fopen(recdcm3tree, "w");
    if(!fp) {
      printf("Error: could not open filename \"%s\"\n", recdcm3tree);
      cleanup(1);
      exit(EXIT_FAILURE);
    }
    fprintf(fp, "%s", newick_tree); fclose(fp);
    /**********/

    del_phylogenetic_tree(T);
    del_phylogenetic_tree(U);

    /*do iterative improvement on Rec-DCM3 tree*/
    if(OUT) fprintf(OUT, "Iterative improvement on Recursive-DCM3 tree...\n");
    else    printf("Iterative improvement on Recursive-DCM3 tree...\n");
    T = global_search(globalcmd, num_to_name, name_to_num, full_taxa, &treescore, scorefile, globalsearchtree);
    for(i=0;i<strlen(newick_tree);i++) newick_tree[i]='\0';
    print_tree(T, num_to_name, newick_tree, 1);
    /*print_rooted_tree(T, num_to_name, newick_tree, 1);*/
    if(OUT){
      fprintf(OUT, "Rec-I-DCM3 tree: %s\n", newick_tree);
      fprintf(OUT, "Rec-I-DCM3 score: %.3f\n", treescore);
    }
    else {
      printf("Rec-I-DCM3 tree: %s\n", newick_tree);
      printf("Rec-I-DCM3 score: %.3f\n", treescore);
    }

    currenttime=time(NULL);
    if (OUT) fprintf(OUT, "Total time (hrs): %.5f\n", difftime(currenttime, starttime)/3600);
    else     printf("Total time (hrs): %.5f\n", difftime(currenttime, starttime)/3600);
  }
  
  cleanup(1);

  /*free memory*/
  if(output_filename) { free(output_filename); fclose(OUT); }
  del_phylogenetic_tree(T);
  for(i=0;i<ntax;i++)free(sequences[i]); free(sequences);
  for(i=0;i<ntax;i++)free(num_to_name[i]); free(num_to_name);
  del_hashtable(name_to_num);
  del_int_set(full_taxa);
  /*******/
  
  return 1;
}

void cleanup(int dummy){

  char* cleanup_cmd; 
  int result;

  /*remove files*/
  if(!DEBUG){
    cleanup_cmd=(char*)malloc(MAX_COMMAND * sizeof(char));
    cleanup_cmd[0]='\0'; strcat(cleanup_cmd,"/bin/rm ");
    strcat(cleanup_cmd, recdcm3tree); strcat(cleanup_cmd, " ");
    strcat(cleanup_cmd, restricted_gt); strcat(cleanup_cmd, " ");
    strcat(cleanup_cmd, subproblemfile); strcat(cleanup_cmd, " ");
    strcat(cleanup_cmd, subtreefile); strcat(cleanup_cmd, " ");
    strcat(cleanup_cmd, globalsearchtree); strcat(cleanup_cmd, " ");
    strcat(cleanup_cmd, scorefile); strcat(cleanup_cmd, " ");
    result = system(cleanup_cmd);
    if(result != 0) {
      printf("Error: could not execute system command \"%s\"\n", cleanup_cmd);
      exit(EXIT_FAILURE);
    }
    free(cleanup_cmd);
  }
  /*********/
}

 
static void read_chars(FILE* fp, char*** num_to_names_, char*** sequences_,
		int* ntax, int* nchars){
  
  int i; char** sequences; char** num_to_names;

  *ntax=0; *nchars=0;
  fscanf(fp, "%d", ntax);
  fscanf(fp, "%d", nchars);
  if(*ntax<=0 || *nchars<=0){
    printf("Error: number of taxa=%d and number of sites=%d must be positive", *ntax, *nchars);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  
  sequences=(char**) malloc(*ntax * sizeof(char*));
  num_to_names=(char**) malloc(*ntax * sizeof(char*));
  for(i=0;i<*ntax;i++){
    sequences[i] = (char*) malloc((*nchars+1) * sizeof(char));
    num_to_names[i] = (char*) malloc(MAX_NAME_LEN * sizeof(char));
    fscanf(fp, "%s", num_to_names[i]);
    fscanf(fp, "%s", sequences[i]);
  }

  /*
  for(i=0;i<*ntax;i++)
    printf("%s\t%s\n", num_to_names[i], sequences[i]);
  */

  *sequences_ = sequences;
  *num_to_names_ = num_to_names;
}

